from typing import Dict, List
import asf_search as asf
from asf_search import ASFSearchResults
import defusedxml.ElementTree as DefusedETree
import xml.etree.ElementTree as ETree
import json
import shapely.wkt as WKT
import requests
import csv

from shapely.geometry import Polygon
from shapely.wkt import loads
from shapely.ops import transform
from shapely.geometry import shape
from shapely.geometry.base import BaseGeometry
from asf_search.constants import PLATFORM

# when this replaces SearchAPI change values to cached
API_URL = 'https://api.daac.asf.alaska.edu/services/search/param?'

def run_test_output_format(results: ASFSearchResults):
    #search results are always sorted this way when returned from asf_search.search(), 
    # but not all test case resources are
    results.sort(key=lambda p: (p.properties['stopTime'], p.properties['fileID']), reverse=True)
    product_list_str = ','.join([product.properties['fileID'] for product in results])

    results.searchComplete = True
    
    for output_type in ['csv', 'kml', 'metalink', 'jsonlite', 'jsonlite2', 'geojson']:
        expected = get_SearchAPI_Output(product_list_str, output_type)
        if output_type == 'csv':
            check_csv(results, expected)
        if output_type == 'kml':
            check_kml(results, expected)
        elif output_type == 'metalink':
            check_metalink(results, expected)
        elif output_type in ['jsonlite', 'jsonlite2']:
            check_jsonLite(results, expected, output_type)
        elif output_type == 'geojson':
            check_geojson(results)

def check_metalink(results: ASFSearchResults, expected_str: str):
    actual = ''.join([line for line in results.metalink()])
    
    actual_tree = DefusedETree.fromstring(actual)
    expected_tree = DefusedETree.fromstring(expected_str)
    
    canon_actual = ETree.canonicalize(DefusedETree.tostring(actual_tree), strip_text=True)
    canon_expected = ETree.canonicalize(DefusedETree.tostring(expected_tree), strip_text=True)
    
    assert canon_actual == canon_expected

def check_kml(results: ASFSearchResults, expected_str: str):
    namespaces = {'kml': 'http://www.opengis.net/kml/2.2'}
    placemarks_path = ".//kml:Placemark"
    expected_root = DefusedETree.fromstring(expected_str)
    expected_placemarks = expected_root.findall(placemarks_path, namespaces)

    actual_root = DefusedETree.fromstring(''.join([block for block in results.kml()]))
    actual_placemarks = actual_root.findall(placemarks_path, namespaces)
    
    # Check polygons for equivalence (asf-search starts from a different pivot)
    # and remove them from the kml so we can easily compare the rest of the placemark data
    for expected_placemark, actual_placemark in zip(expected_placemarks, actual_placemarks):
        expected_polygon = expected_placemark.findall('./*')[-1]
        actual_polygon = actual_placemark.findall('./*')[-1]
        
        expected_coords = get_coordinates_from_kml(DefusedETree.tostring(expected_polygon))
        actual_coords = get_coordinates_from_kml(DefusedETree.tostring(actual_polygon))
        
        assert Polygon(expected_coords).equals(Polygon(actual_coords))
        
        expected_placemark.remove(expected_polygon)
        actual_placemark.remove(actual_polygon)
        
    # Get canonicalize xml strings so minor differences are normalized
    actual_canon = ETree.canonicalize( DefusedETree.tostring(actual_root), strip_text=True)
    expected_canon = ETree.canonicalize( DefusedETree.tostring(expected_root), strip_text=True)
    
    assert actual_canon == expected_canon


def get_coordinates_from_kml(data: str):
    namespaces = {'kml': 'http://www.opengis.net/kml/2.2'}

    coords = []
    coords_lon_lat_path = ".//kml:outerBoundaryIs/kml:LinearRing/kml:coordinates"
    root = DefusedETree.fromstring(data)
    
    coordinates_elements = root.findall(coords_lon_lat_path, namespaces)
    for lon_lat_z in coordinates_elements[0].text.split('\n'):
        if len(lon_lat_z.split(',')) == 3:
            lon, lat, _ = lon_lat_z.strip().split(',')
            coords.append([float(lon), float(lat)])

    return coords
    

def check_csv(results: ASFSearchResults, expected_str: str):
    expected = [product for product in csv.reader(expected_str.split('\n')) if product != []]
    # actual = [prod for prod in csv.reader(''.join([s for s in results.csv()]).split('\n')) if prod != []]
    
    expected = csv.DictReader(expected_str.split('\n'))
    actual = csv.DictReader([s for s in results.csv()])
    
    for actual_row, expected_row in zip(actual, expected):
        actual_dict = dict(actual_row)
        expected_dict = dict(expected_row)
        
        for key in expected_dict.keys():
            if expected_dict[key] in ['None', None, '']:
                assert actual_dict[key] in ['None', None, '']
            else:
                try:
                    expected_value = float(expected_dict[key])
                    actual_value = float(actual_dict[key])
                    assert expected_value == actual_value, \
                        f"expected \'{expected_dict[key]}\' for key \'{key}\', got \'{actual_dict[key]}\'"
                except ValueError:
                    assert expected_dict[key] == actual_dict[key], \
                        f"expected \'{expected_dict[key]}\' for key \'{key}\', got \'{actual_dict[key]}\'"
 
 
def check_jsonLite(results: ASFSearchResults, expected_str: str, output_type: str):
    jsonlite2 = output_type == 'jsonlite2'
    
    expected = json.loads(expected_str)['results']

    if jsonlite2:
        wkt_key = 'w'
        wkt_unwrapped_key = 'wu'
    else:
        wkt_key = 'wkt'
        wkt_unwrapped_key = 'wkt_unwrapped'

    actual = json.loads(''.join(results.jsonlite2() if jsonlite2 else results.jsonlite()))['results']

    for idx, expected_product in enumerate(expected):
        wkt = expected_product.pop(wkt_key)
        wkt_unwrapped = expected_product.pop(wkt_unwrapped_key)
        
        for key in expected_product.keys():
            assert actual[idx][key] == expected_product[key]
        
        assert WKT.loads(actual[idx][wkt_key]).equals(WKT.loads(wkt))
        assert WKT.loads(actual[idx][wkt_unwrapped_key]).equals(WKT.loads(wkt_unwrapped))

def check_geojson(results: ASFSearchResults):
    expected = results.geojson()
    actual = asf.export.results_to_geojson(results)
    
    assert json.loads(''.join(actual)) == expected
    
def get_SearchAPI_Output(product_list: List[str], output_type: str) -> List[Dict]:
    response = requests.get(API_URL, [('product_list', product_list), ('output', output_type)])
    response.raise_for_status()
    
    expected = response.text
    
    return expected

def run_test_ASFSearchResults_intersection(wkt: str):
    wrapped, unwrapped, _ = asf.validate_wkt(wkt)
    unchanged_aoi = loads(wkt) # sometimes geometries don't come back with wrapping in mind

    # exclude SMAP products
    platforms = [
                 PLATFORM.ALOS,
                 PLATFORM.SENTINEL1,
                 PLATFORM.SIRC, 
                 PLATFORM.UAVSAR
                 ]
    
    def overlap_check(s1: BaseGeometry, s2: BaseGeometry):
        return s1.overlaps(s2) or s1.touches(s2) or s2.distance(s1) <= 0.005

    for platform in platforms:
        results = asf.geo_search(intersectsWith=wkt, platform=platform, maxResults=250)

        for product in results:
            if shape(product.geometry).is_valid:
                product_geom_wrapped, product_geom_unwrapped, _ = asf.validate_wkt(shape(product.geometry))
                original_shape = unchanged_aoi

                assert overlap_check(product_geom_wrapped, wrapped) or overlap_check(product_geom_wrapped, original_shape), f"OVERLAP FAIL: {product.properties['sceneName']}, {product.geometry} \nproduct: {product_geom_wrapped.wkt} \naoi: {wrapped.wkt}"



from typing import List
import itertools
from copy import copy

from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.constants import CMR_PAGE_SIZE


def build_subqueries(opts: ASFSearchOptions) -> List[ASFSearchOptions]:
    """
    Build a list of sub-queries using the cartesian product of all the list parameters described by opts

    :param opts: The search options to split into sub-queries

    :return list: A list of ASFSearchOptions objects
    """
    params = dict(opts)

    # Break out two big list offenders into manageable chunks
    if params.get('granule_list') is not None:
        params['granule_list'] = chunk_list(params['granule_list'], CMR_PAGE_SIZE)
    if params.get('product_list') is not None:
        params['product_list'] = chunk_list(params['product_list'], CMR_PAGE_SIZE)

    list_param_names = ['platform', 'season', 'collections', 'dataset']  # these parameters will dodge the subquery system
    skip_param_names = ['maxResults']# these params exist in opts, but shouldn't be passed on to subqueries at ALL
    
    params = dict([ (k, v) for k, v in params.items() if k not in skip_param_names ])

    subquery_params, list_params = {}, {}
    for k, v in params.items():
        if k in list_param_names:
            list_params[k] = v
        else:
            subquery_params[k] = v

    sub_queries = cartesian_product(subquery_params)

    final_sub_query_opts = []
    for query in sub_queries:
        q = dict()
        for p in query:
            q.update(p)
        q['provider'] = opts.provider
        q['host'] = opts.host
        q['session'] = copy(opts.session)
        for key in list_params.keys():
            q[key] = list_params[key]

        final_sub_query_opts.append(ASFSearchOptions(**q))

    return final_sub_query_opts


def chunk_list(source: List, n: int) -> List:
    """
    Breaks a longer list into a list of lists, each of length n

    :param source: The list to be broken into chunks
    :param n: The maximum length of each chunk

    :return List[List, ...]:
    """
    return [source[i * n:(i + 1) * n] for i in range((len(source) + n - 1) // n)]


def cartesian_product(params):
    formatted_params = format_query_params(params)
    p = list(itertools.product(*formatted_params))
    return p


def format_query_params(params):
    listed_params = []

    for param_name, param_val in params.items():
        plist = translate_param(param_name, param_val)
        listed_params.append(plist)

    return listed_params


def translate_param(param_name, param_val):
    param_list = []

    if not isinstance(param_val, list):
        param_val = [param_val]

    for unformatted_val in param_val:
        formatted_val = unformatted_val

        if isinstance(unformatted_val, list):
            formatted_val = ','.join([f'{t}' for t in unformatted_val])

        param_list.append({
            param_name: formatted_val
        })

    return param_list



import inspect
from types import GeneratorType
import xml.etree.ElementTree as ETree
from asf_search import ASF_LOGGER
from asf_search.export.export_translators import ASFSearchResults_to_properties_list

def results_to_metalink(results):
    ASF_LOGGER.info('Started translating results to metalink format')
    
    if inspect.isgeneratorfunction(results) or isinstance(results, GeneratorType):
        return MetalinkStreamArray(results)
    
    return MetalinkStreamArray([results])

class MetalinkStreamArray(list):
    def __init__(self, results):
        self.pages = results
        self.len = 1
        self.header = """<?xml version="1.0"?>
<metalink xmlns="http://www.metalinker.org/" version="3.0">
    <publisher><name>Alaska Satellite Facility</name><url>http://www.asf.alaska.edu/</url></publisher>
    <files>"""

        self.footer = """
    </files>\n</metalink>"""

    def get_additional_fields(self, product):
        return {}
    
    def __iter__(self):
        return self.streamPages()

    def __len__(self):
        return self.len

    def streamPages(self):
        yield self.header
        
        completed = False
        for page_idx, page in enumerate(self.pages):
            ASF_LOGGER.info(f"Streaming {len(page)} products from page {page_idx}")
            completed = page.searchComplete
            
            properties_list = ASFSearchResults_to_properties_list(page, self.get_additional_fields)
            yield from [self.getItem(p) for p in properties_list]
        
        if not completed:
            ASF_LOGGER.warn('Failed to download all results from CMR')
        
        yield self.footer
        
        ASF_LOGGER.info(f"Finished streaming {self.getOutputType()} results")
        

    def getOutputType(self) -> str:
        return 'metalink'
    
    def getItem(self, p):
        file = ETree.Element("file", attrib={'name': p['fileName']})
        resources = ETree.Element('resources')

        url = ETree.Element('url', attrib={'type': 'http'})
        url.text = p['url']
        resources.append(url)
        file.append(resources)
        
        if p['md5sum'] and p['md5sum'] != 'NA':
            verification = ETree.Element('verification')
            h = ETree.Element('hash', {'type': 'md5'})
            h.text = p['md5sum']
            verification.append(h)
            file.append(verification)
            
        if p['bytes'] and p['bytes'] != 'NA':
            size = ETree.Element('size')
            size.text = str(p['bytes'])
            file.append(size)
        
        return '\n' + (8*' ') + ETree.tostring(file, encoding='unicode')

    def indent(self, elem, level=0):
        # Only Python 3.9+ has a built-in indent function for element tree.
        # https://stackoverflow.com/a/33956544
        i = "\n" + level*"  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self.indent(elem, level+1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i


from types import FunctionType
from datetime import datetime

from asf_search import ASFSearchResults

# ASFProduct.properties don't have every property required of certain output formats, 
# This grabs the missing properties from ASFProduct.umm required by the given format
def ASFSearchResults_to_properties_list(results: ASFSearchResults, get_additional_fields: FunctionType):
    property_list = []
    
    for product in results:
        additional_fields = get_additional_fields(product)
        properties = {**product.properties, **additional_fields}
        property_list.append(properties)
    
    # Format dates to match format used by SearchAPI output formats
    for product in property_list:
        # S1 date properties are formatted differently from other platforms
        is_S1 = product['platform'].upper() in ['SENTINEL-1', 'SENTINEL-1B', 'SENTINEL-1A']
        for key, data in product.items():
            if ('date' in key.lower() or 'time' in key.lower()) and data is not None:
                if not is_S1:
                    # Remove trailing zeroes from miliseconds, add Z
                    if len(data.split('.')) == 2:
                        d = len(data.split('.')[0])
                        data = data[:d] + 'Z'
                    time = datetime.strptime(data, '%Y-%m-%dT%H:%M:%SZ')
                    product[key] = time.strftime('%Y-%m-%dT%H:%M:%SZ')

    return property_list



from typing import Dict, List
from asf_search import ASFSearchOptions, ASFProduct, FileDownloadType
from asf_search.exceptions import ASFAuthenticationError

from ASFProduct.test_ASFProduct import run_test_ASFProduct, run_test_ASFProduct_download, run_test_product_get_stack_options, run_test_stack
from ASFSearchOptions.test_ASFSearchOptions import run_test_ASFSearchOptions
from ASFSearchResults.test_ASFSearchResults import run_test_output_format, run_test_ASFSearchResults_intersection
from ASFSession.test_ASFSession import run_auth_with_cookiejar, run_auth_with_creds, run_auth_with_token, run_test_asf_session_rebuild_auth
from BaselineSearch.test_baseline_search import *
from Search.test_search import run_test_ASFSearchResults, run_test_dataset_search, run_test_search, run_test_search_http_error
from Search.test_search_generator import run_test_search_generator, run_test_search_generator_multi
from CMR.test_MissionList import run_test_get_project_names


from pytest import raises
from unittest.mock import patch

import os
import pathlib
import yaml

from WKT.test_validate_wkt import run_test_search_wkt_prep, run_test_validate_wkt_get_shape_coords, run_test_validate_wkt_clamp_geometry, run_test_validate_wkt_valid_wkt, run_test_validate_wkt_convex_hull, run_test_validate_wkt_counter_clockwise_reorientation, run_test_validate_wkt_invalid_wkt_error, run_test_validate_wkt_merge_overlapping_geometry, run_test_simplify_aoi
from ASFSearchOptions.test_ASFSearchOptions import run_test_ASFSearchOptions_validator, run_test_validator_map_validate
from BaselineSearch.Stack.test_stack import run_test_find_new_reference, run_test_get_baseline_from_stack, run_test_get_default_product_type, run_test_valid_state_vectors

from download.test_download import run_test_download_url, run_test_download_url_auth_error
from Serialization.test_serialization import run_test_serialization
import nbformat
from nbconvert.preprocessors import ExecutePreprocessor


# asf_search.ASFProduct Tests
def test_ASFProduct(**args) -> None:
    """
    Tests Basic ASFProduct with mock searchAPI response
    """
    test_info = args["test_info"]
    geographic_response = get_resource(test_info["products"])
    run_test_ASFProduct(geographic_response)

def test_ASFProduct_Stack(**args) -> None:
    """
    Tests ASFProduct.stack() with reference and corresponding stack
    Checks for temporalBaseline order, 
    asserting the stack is ordered by the scene's temporalBaseline (in ascending order)
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    preprocessed_stack = get_resource(test_info["preprocessed_stack"])
    processed_stack = get_resource(test_info["processed_stack"])
    run_test_stack(reference, preprocessed_stack, processed_stack)

def test_ASFProduct_get_stack_options(**args) -> None:
    test_info = args["test_info"]
    reference = get_resource(test_info['product'])
    options = get_resource(test_info['options'])

    run_test_product_get_stack_options(reference, options)

def test_ASFProduct_download(**args) -> None:
    test_info = args["test_info"]
    reference = get_resource(test_info['product'])
    filename = test_info['filename']
    filetype_raw = test_info['filetype']
    additional_urls = test_info['additionalUrls']

    if filetype_raw == 1:
        filetype = FileDownloadType.DEFAULT_FILE
    elif filetype_raw == 2:
        filetype = FileDownloadType.ADDITIONAL_FILES
    else:
        filetype = FileDownloadType.ALL_FILES
    
    run_test_ASFProduct_download(reference, filename, filetype, additional_urls)
    
# asf_search.ASFSession Tests
def test_ASFSession_Error(**args) -> None:
    """
    Test ASFSession.auth_with_creds for sign in errors
    """
    test_info = args["test_info"]
    username = test_info["username"]
    password = test_info["password"]
    with patch('asf_search.ASFSession.get') as mock_get:
        mock_get.return_value = "Error"

        with raises(ASFAuthenticationError):
            run_auth_with_creds(username, password)

def test_ASFSession_Token_Error(**args) -> None:
    """
    Test ASFSession.auth_with_token for sign in errors
    """
    test_info = args["test_info"]
    token = test_info["token"]

    with raises(ASFAuthenticationError):
        run_auth_with_token(token)

def test_ASFSession_Cookie_Error(**args) -> None:
    """
    Test ASFSession.auth_with_cookie for sign in errors
    """
    test_info = args["test_info"]
    cookies = test_info["cookies"]

    with raises(ASFAuthenticationError):
        run_auth_with_cookiejar(cookies)

def test_asf_session_rebuild_auth(**args) -> None:
    """
    Test asf_search.ASFSession.rebuild_auth
    When redirecting from an ASF domain, only accept 
    domains listed in ASFSession.AUTH_DOMAINS
    """
    test_info = args["test_info"]
    original_domain = test_info["original_domain"]
    response_domain = test_info["response_domain"]
    response_code = test_info["response_code"]
    final_token = test_info["final_token"]

    run_test_asf_session_rebuild_auth(original_domain, response_domain, response_code, final_token)

# asf_search.search.baseline_search Tests
def test_get_preprocessed_stack_params(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_opts with a reference scene
    that's part of a pre-calculated platform, asserting that get_stack_opts returns an object with two parameters
    \n1. processingLevel
    \n2. insarStackId
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])

    run_test_get_preprocessed_stack_params(reference)

def test_get_unprocessed_stack_params(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_opts with a reference scene
    that's not part of a pre-calculated platform, asserting that get_stack_opts returns an object with seven parameters
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])

    run_test_get_unprocessed_stack_params(reference)

def test_get_stack_opts_invalid_insarStackId(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_opts with a the reference scene's 
    insarStackID set to an invalid value, and asserting an ASFBaselineError is raised
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    
    run_get_stack_opts_invalid_insarStackId(reference)
    
def test_get_stack_opts_invalid_platform(**args) -> None:
    """
    Test asf_search.search.baseline_search.get_stack_opts with a the reference scene's 
    platform set to an invalid value, and asserting an ASFBaselineError is raised
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])    
    run_test_get_stack_opts_invalid_platform_raises_error(reference)
    
def test_temporal_baseline(**args) -> None:
    """
    Test asf_search.search.baseline_search.calc_temporal_baselines, asserting mutated baseline stack
    is still the same length and that each product's properties contain a temporalBaseline key 
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    stack = get_resource(test_info["stack"])
    run_test_calc_temporal_baselines(reference, stack)
    
def test_stack_from_product(**args) -> None:
    """
    Test asf_search.search.baseline_search.stack_from_product, asserting stack returned is ordered
    by temporalBaseline value in ascending order
    """
    test_info = args["test_info"]
    reference = get_resource(test_info["product"])
    stack = get_resource(test_info["stack"])
    
    run_test_stack_from_product(reference, stack)

def test_stack_from_id(**args) -> None:
    """
    Test asf_search.search.baseline_search.stack_from_id, asserting stack returned is ordered
    by temporalBaseline value in ascending order
    """
    test_info = args["test_info"]
    stack_id = test_info["stack_id"]
    stack_reference_data = test_info["stack_reference"]
    stack_data = test_info["stack"]

    stack_reference = get_resource(stack_reference_data)
    stack = []

    if(stack_data != []):
        stack = get_resource(stack_data)

    run_test_stack_from_id(stack_id, stack_reference, stack)

# asf_search.ASFSearchResults Tests
def test_ASFSearchResults(**args) -> None:
    """
    Test asf_search.ASFSearchResults, asserting initialized values, 
    and geojson response returns object with type FeatureCollection
    """
    test_info = args["test_info"]
    search_response = get_resource(test_info["response"])

    run_test_ASFSearchResults(search_response)

# asf_search.search Tests
def test_ASFSearch_Search(**args) -> None:
    """
    Test asf_search.search, asserting returned value is expected result
    """
    test_info = args["test_info"]
    parameters = get_resource(test_info["parameters"])
    answer = get_resource(test_info["answer"])

    run_test_search(parameters, answer)

def test_ASFSearch_Search_Generator(**args) -> None:
    test_info = args["test_info"]
    params = get_resource(test_info['parameters'])

    if isinstance(params, list):
        opts = []
        for p in params:
            opts.append(ASFSearchOptions(**p))

        run_test_search_generator_multi(opts)

    else:
        run_test_search_generator(ASFSearchOptions(**params))


def test_ASFSearch_Search_Error(**args) -> None:
    """
    Test asf_search.search errors,
    asserting server and client errors are raised
    """
    test_info = args["test_info"]
    parameters = test_info["parameters"]
    report = test_info["report"]
    error_code = test_info["status_code"]

    run_test_search_http_error(parameters, error_code, report)

def test_wkt_validation_Invalid_WKT_Error(**args) -> None:
    """
    Test asf_search.wkt errors,
    asserting wkt validation errors are raised
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    run_test_validate_wkt_invalid_wkt_error(wkt)

def test_wkt_validation_WKT_Valid(**args) -> None:
    """
    Test asf_search.validate_wkt, asserting expected wkts are returned
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    validated_wkt = get_resource(test_info['validated-wkt'])
    run_test_validate_wkt_valid_wkt(wkt, validated_wkt)

def test_wkt_validation_WKT_clamp_geometry(**args) -> None:
    """
    Test asf_search.validate_wkt._get_clamped_and_wrapped_geometry, asserting the amount of clamped and wrapped coordinates
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    clamped_wkt = get_resource(test_info['clamped-wkt'])
    clamped_count = get_resource(test_info['clamped-count'])
    wrapped_count = get_resource(test_info['wrapped-count'])
    run_test_validate_wkt_clamp_geometry(wkt, clamped_wkt, clamped_count, wrapped_count)

def test_wkt_validation_convex_hull(**args) -> None:
    """
    Test asf_search.validate_wkt._get_convex_hull, asserting convex hulls producted are expected
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    convex_wkt = get_resource(test_info['convex-wkt'])
    run_test_validate_wkt_convex_hull(wkt, convex_wkt)

def test_wkt_validation_merge_overlapping_geometry(**args) -> None:
    """
    Test asf_search.validate_wkt._merge_overlapping_geometry, asserting expected shapes are merged
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    merged_wkt = get_resource(test_info['merged-wkt'])
    run_test_validate_wkt_merge_overlapping_geometry(wkt, merged_wkt)

def test_wkt_validation_counter_clockwise_reorientation(**args) -> None:
    """
    Test asf_search.validate_wkt._counter_clockwise_reorientation reverses polygon orientation if polygon is wound clockwise,
    and maintains counter-clockwise winding when polygon orientation is correct
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    cc_wkt = get_resource(test_info['cc-wkt'])
    run_test_validate_wkt_counter_clockwise_reorientation(wkt, cc_wkt)

def test_validate_wkt_get_shape_coords(**args) -> None:
    """
    Test asf_search.validate_wkt._get_shape_coords asserting all coordinates are returned and expected
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    coords = get_resource(test_info['coordinates'])
    run_test_validate_wkt_get_shape_coords(wkt, coords)

def test_search_wkt_prep(**args) -> None:
    """
    Test asf_search.validate_wkt.wkt_prep, asserting returned shape is correct geometric type and expected shape
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    
    run_test_search_wkt_prep(wkt)

def test_simplify_aoi(**args) -> None:
    """
    Test asf_search.validate_wkt.wkt_prep, asserting returned shape is correct geometric type and expected shape
    """
    test_info = args["test_info"]
    wkt = get_resource(test_info['wkt'])
    simplified = get_resource(test_info["simplified-wkt"])
    RepairEntries = get_resource(test_info["RepairEntries"])
    run_test_simplify_aoi(wkt, simplified, RepairEntries)

def test_get_platform_campaign_names(**args) -> None:
    test_info = args["test_info"]
    cmr_ummjson = get_resource(test_info["cmr_ummjson"])
    campaigns: List[str] = get_resource(test_info["campaigns"])
    
    run_test_get_project_names(cmr_ummjson, campaigns)

def test_download_url(**args) -> None:
    """
    Test asf_search.download.download_url
    """
    test_info = args["test_info"]
    url = test_info["url"]
    path = test_info["path"]
    filename = test_info["filename"]

    if filename == "error":
        run_test_download_url_auth_error(url, path, filename)
    else:
        run_test_download_url(url, path, filename)
        
def test_find_new_reference(**args) -> None:
    """
    Test asf_search.baseline.calc.find_new_reference
    """
    test_info = args["test_info"]
    stack = get_resource(test_info["stack"])
    output_index = get_resource(test_info["output_index"])
    
    run_test_find_new_reference(stack, output_index)

def test_get_default_product_type(**args) -> None:
    test_info = args["test_info"]
    product = get_resource(test_info["product"])
    product_type = get_resource(test_info["product_type"])
    
    product = ASFProduct(args={'meta': product['meta'], 'umm': product['umm']})
    if product.properties.get('sceneName') is None:
        product.properties['sceneName'] = 'BAD_SCENE'
        
    run_test_get_default_product_type(product, product_type)

def test_get_baseline_from_stack(**args) -> None:
    test_info = args["test_info"]
    reference = get_resource(test_info['reference'])
    stack = get_resource(test_info['stack'])
    output_stack = get_resource(test_info['output_stack'])
    error = get_resource(test_info['error'])
    run_test_get_baseline_from_stack(reference, stack, output_stack, error)

def test_valid_state_vectors(**args) -> None:
    test_info = args["test_info"]
    reference = get_resource(test_info['reference'])
    output = get_resource(test_info['output'])
    
    run_test_valid_state_vectors(reference, output)
    
def test_validator_map_validate(**args) -> None:
    test_info = args["test_info"]
    key = get_resource(test_info['key'])
    value = get_resource(test_info['value'])
    output = get_resource(test_info['output'])

    run_test_validator_map_validate(key, value, output)

def test_ASFSearchOptions_validator(**args) -> None:
    test_info = args["test_info"]
    validator_name = get_resource(test_info['validator'])
    param = safe_load_tuple(get_resource(test_info['input']))
    output = safe_load_tuple(get_resource(test_info['output']))
    error = get_resource(test_info['error'])
    run_test_ASFSearchOptions_validator(validator_name, param, output, error)
    

def test_ASFSearchOptions(**kwargs) -> None:
    run_test_ASFSearchOptions(**kwargs)

def test_ASFSearchResults_intersection(**kwargs) -> None:
    wkt = get_resource(kwargs['test_info']['wkt'])
    run_test_ASFSearchResults_intersection(wkt)

def test_search_dataset(**kwargs) -> None:
    dataset = get_resource(kwargs['test_info']['dataset'])
    run_test_dataset_search(dataset)
    
def test_serialization(**args) -> None:
    test_info = args['test_info']
    product = get_resource(test_info.get('product'))
    results = get_resource(test_info.get('results'))
    search_opts = get_resource(test_info.get('searchOpts'))
    options = ASFSearchOptions(**search_opts if search_opts else {})

    run_test_serialization(product, results, options)

def test_notebook_examples(**args) -> None:
    test_info = args['test_info']
    notebook_file = test_info['notebook']
    path = os.path.join('examples', notebook_file)

    with open(path) as f:
        notebook = nbformat.read(f, as_version=4)
        ep = ExecutePreprocessor(timeout=600)
        try:
            assert ep.preprocess(notebook) != None, f"Got empty notebook for {notebook_file}"
        except Exception as e:
            assert False, f"Failed executing {notebook_file}: {e}"


# Testing resource loading utilities

def safe_load_tuple(param):
    """
    loads a tuple from a list if a param is an object with key 'tuple'
    (Arbritrary constructor initialization is not supported by yaml.safe_load
    as a security measure)
    
    """
    if isinstance(param, Dict):
        if "tuple" in param.keys():
            param = tuple(param['tuple'])
    
    return param

def test_output_format(**args) -> None:
    test_info = args['test_info']
    
    products = get_resource(test_info['results'])
    if not isinstance(products, List):
        products = [products]
    results = ASFSearchResults([ASFProduct(args={'meta': product['meta'], 'umm': product['umm']}) for product in products])

    run_test_output_format(results)

# Finds and loads file from yml_tests/Resouces/ if loaded field ends with .yml/yaml extension
def get_resource(yml_file):
    
    if isinstance(yml_file, str):
        if yml_file.endswith((".yml", ".yaml")):
            base_path = pathlib.Path(__file__).parent.resolve()
            with open(os.path.join(base_path, "yml_tests", "Resources", yml_file), "r") as f:
                try:
                    return yaml.safe_load(f)
                except yaml.YAMLError as exc:
                    print(exc)
    elif isinstance(yml_file, List): #check if it's a list of yml files
        if len(yml_file) > 0:
            if isinstance(yml_file[0], str):
                if yml_file[0].endswith((".yml", ".yaml")):
                    return [get_resource(file) for file in yml_file]

    return yml_file



from .MissionList import get_campaigns
from .subquery import build_subqueries
from .translate import translate_product, translate_opts, get_additional_fields
from .field_map import field_map
from .datasets import dataset_collections


from asf_search import ASF_LOGGER

from .validators import (
    parse_string, parse_float, parse_wkt, parse_date,
    parse_string_list, parse_int_list, parse_int_or_range_list,
    parse_float_or_range_list,
    parse_session
)


def validate(key, value):
    if key not in validator_map:
        error_msg = f"Key '{key}' is not a valid search option."
        # See if they just missed up case sensitivity:
        for valid_key in validator_map:
            if key.lower() == valid_key.lower():
                error_msg += f" (Did you mean '{valid_key}'?)"
                break
        ASF_LOGGER.error(error_msg)
        raise KeyError(error_msg)
    try:
        return validator_map[key](value)
    except ValueError as exc:
        ASF_LOGGER.exception(f"Failed to parse item in ASFSearchOptions: {key=} {value=} {exc=}")
        raise

validator_map = {
    # Search parameters       Parser
    'maxResults':             int,
    'absoluteOrbit':          parse_int_or_range_list,
    'asfFrame':               parse_int_or_range_list,
    'beamMode':               parse_string_list,
    'beamSwath':              parse_string_list,
    'campaign':               parse_string,
    'maxDoppler':             parse_float,
    'minDoppler':             parse_float,
    'maxFaradayRotation':     parse_float,
    'minFaradayRotation':     parse_float,
    'flightDirection':        parse_string,
    'flightLine':             parse_string,
    'frame':                  parse_int_or_range_list,
    'granule_list':           parse_string_list,
    'product_list':           parse_string_list,
    'intersectsWith':         parse_wkt,
    'lookDirection':          parse_string,
    'offNadirAngle':          parse_float_or_range_list,
    'platform':               parse_string_list,
    'polarization':           parse_string_list,
    'processingLevel':        parse_string_list,
    'relativeOrbit':          parse_int_or_range_list,
    'processingDate':         parse_date,
    'start':                  parse_date,
    'end':                    parse_date,
    'season':                 parse_int_list,
    'groupID':                parse_string_list,
    'insarStackId':           parse_string,
    'instrument':             parse_string,
    'collections':            parse_string_list,
    'temporalBaselineDays':   parse_string_list,
    'operaBurstID':           parse_string_list,
    'absoluteBurstID':        parse_int_list,
    'relativeBurstID':        parse_int_list,
    'fullBurstID':            parse_string_list,
    'dataset':                parse_string_list,

    # Config parameters       Parser
    'session':                parse_session,
    'host':                   parse_string,
    'provider':               parse_string
}



from .search import search
from .granule_search import granule_search
from .product_search import product_search
from .geo_search import geo_search
from .baseline_search import stack_from_id, get_stack_opts
from .campaigns import campaigns
from .search_count import search_count
from .search_generator import search_generator


SENTINEL1 = 'SENTINEL-1'
OPERA_S1 = 'OPERA-S1'
SLC_BURST = 'SLC-BURST'
ALOS_PALSAR = 'ALOS PALSAR'
ALOS_AVNIR_2 = 'ALOS AVNIR-2'
SIRC = 'SIR-C'
ARIA_S1_GUNW = 'ARIA S1 GUNW'
SMAP = 'SMAP'
UAVSAR = 'UAVSAR'
RADARSAT_1 = 'RADARSAT-1'
ERS = 'ERS'
JERS_1 = 'JERS-1'
AIRSAR = 'AIRSAR'
SEASAT = 'SEASAT'


from copy import deepcopy
from unittest.mock import patch
from asf_search.exceptions import ASFBaselineError, ASFSearchError
from asf_search.ASFSearchResults import ASFSearchResults
from asf_search import ASFProduct
from asf_search.search.baseline_search import get_stack_opts, stack_from_id, stack_from_product
from asf_search.baseline.stack import calculate_temporal_baselines, get_default_product_type
import pytest

def run_test_get_preprocessed_stack_params(product):
    reference = ASFProduct(product)
    params = get_stack_opts(reference)

    original_properties = product['properties']
    
    assert(params.processingLevel == [get_default_product_type(reference)])
    assert(params.insarStackId == original_properties['insarStackId'])
    assert(len(dict(params)) == 2)
    

def run_test_get_unprocessed_stack_params(product):
    reference = ASFProduct(product)
    params = get_stack_opts(reference)

    original_properties = product['properties']
    assert(original_properties['polarization'] in params.polarization)
    
    if reference.properties['processingLevel'] == 'BURST':
        assert([reference.properties['polarization']] == params.polarization)
        assert([reference.properties['burst']['fullBurstID']] == params.fullBurstID)
    else:
        assert(['VV', 'VV+VH'] == params.polarization if reference.properties['polarization'] in ['VV', 'VV+VH'] else ['HH','HH+HV'] == params.polarization)
        assert(len(dict(params)) == 7)

def run_get_stack_opts_invalid_insarStackId(product):
    invalid_reference = ASFProduct(product)
    
    invalid_reference.properties['insarStackId'] = '0'

    with pytest.raises(ASFBaselineError):
        get_stack_opts(invalid_reference)
    
def run_test_get_stack_opts_invalid_platform_raises_error(product):
    invalid_reference = ASFProduct(product)
    invalid_reference.properties['platform'] = 'FAKE_PLATFORM'
    
    with pytest.raises(ASFBaselineError):
        get_stack_opts(invalid_reference)
    
def run_test_calc_temporal_baselines(reference, stack):
    reference = ASFProduct(reference)
    stack = ASFSearchResults([ASFProduct(product) for product in stack])
    stackLength = len(stack)

    calculate_temporal_baselines(reference, stack)

    assert(len(stack) == stackLength)
    for secondary in stack:
        assert('temporalBaseline' in secondary.properties)

def run_test_stack_from_product(reference, stack):
    reference = ASFProduct(reference)

    with patch('asf_search.baseline_search.search') as search_mock:
        search_mock.return_value = ASFSearchResults([ASFProduct(product) for product in stack])  

        stack = stack_from_product(reference)

        for (idx, secondary) in enumerate(stack):
            if(idx > 0):
                assert(secondary.properties['temporalBaseline'] >= stack[idx - 1].properties['temporalBaseline'])

def run_test_stack_from_id(stack_id: str, reference, stack):
        temp = deepcopy(stack)

        with patch('asf_search.baseline_search.product_search') as mock_product_search:
            mock_product_search.return_value = ASFSearchResults([ASFProduct(product) for product in stack])
        
            if not stack_id:    
                with pytest.raises(ASFSearchError):
                    stack_from_id(stack_id)
            else:
                with patch('asf_search.baseline_search.search') as search_mock:
                    search_mock.return_value = ASFSearchResults([ASFProduct(product) for product in temp])   

                    returned_stack = stack_from_id(stack_id)
                    assert(len(returned_stack) == len(stack))

                    for (idx, secondary) in enumerate(returned_stack):
                        if(idx > 0):
                            assert(secondary.properties['temporalBaseline'] >= stack[idx - 1]["properties"]['temporalBaseline'])



from typing import Iterable
from multiprocessing import Pool
import os.path
from urllib import parse
from requests import Response
from requests.exceptions import HTTPError
import warnings
import regex as re

from asf_search.exceptions import ASFAuthenticationError, ASFDownloadError
from asf_search import ASFSession
from remotezip import RemoteZip
from tenacity import retry, stop_after_delay, retry_if_result, wait_fixed

def _download_url(arg):
    url, path, session = arg
    download_url(
        url=url,
        path=path,
        session=session)


def download_urls(urls: Iterable[str], path: str, session: ASFSession = None, processes: int = 1):
    """
    Downloads all products from the specified URLs to the specified location.

    :param urls: List of URLs from which to download
    :param path: Local path in which to save the product
    :param session: The session to use, in most cases should be authenticated beforehand
    :param processes: Number of download processes to use. Defaults to 1 (i.e. sequential download)
    :return:
    """
    if session is None:
        session = ASFSession()

    if processes <= 1:
        for url in urls:
            download_url(url=url, path=path, session=session)
    else:
        pool = Pool(processes=processes)
        args = [(url, path, session) for url in urls]
        pool.map(_download_url, args)
        pool.close()
        pool.join()


def download_url(url: str, path: str, filename: str = None, session: ASFSession = None) -> None:
    """
    Downloads a product from the specified URL to the specified location and (optional) filename.

    :param url: URL from which to download
    :param path: Local path in which to save the product
    :param filename: Optional filename to be used, extracted from the URL by default
    :param session: The session to use, in most cases should be authenticated beforehand
    :return:
    """
    
    if filename is None:
        filename = os.path.split(parse.urlparse(url).path)[1]
    
    if not os.path.isdir(path):
        raise ASFDownloadError(f'Error downloading {url}: directory not found: {path}')

    if os.path.isfile(os.path.join(path, filename)):
        warnings.warn(f'File already exists, skipping download: {os.path.join(path, filename)}')
        return

    if session is None:
        session = ASFSession()

    response = _try_get_response(session=session, url=url)
    
    with open(os.path.join(path, filename), 'wb') as f:
        for chunk in response.iter_content(chunk_size=8192):
            f.write(chunk)

def remotezip(url: str, session: ASFSession) -> RemoteZip:
    """
    :param url: the url to the zip product
    :param session: the authenticated ASFSession to read and download from the zip file
    """

    session.hooks['response'].append(strip_auth_if_aws)
    return RemoteZip(url, session=session)

def strip_auth_if_aws(r, *args, **kwargs):
    if 300 <= r.status_code <= 399 and 'amazonaws.com' in parse.urlparse(r.headers['location']).netloc:
        location = r.headers['location']
        r.headers.clear()
        r.headers['location'] = location

# if it's an unprocessed burst product it'll return a 202 and we'll have to query again 
# https://sentinel1-burst-docs.asf.alaska.edu/
def _is_burst_processing(response: Response):
    return response.status_code == 202

@retry(reraise=True,
       retry=retry_if_result(_is_burst_processing),
       wait=wait_fixed(1),
       stop=stop_after_delay(90),
    )
def _try_get_response(session: ASFSession, url: str):
    response = session.get(url, stream=True, hooks={'response': strip_auth_if_aws})

    try:
        response.raise_for_status()
    except HTTPError as e:
        if 400 <= response.status_code <= 499:
            raise ASFAuthenticationError(f'HTTP {e.response.status_code}: {e.response.text}')

        raise e

    return response



import logging
from typing import Union, Tuple, List
from shapely import wkt
from shapely.geometry.base import BaseGeometry
from shapely.geometry import Polygon, MultiPolygon, Point, LineString, GeometryCollection
from shapely.geometry.collection import BaseMultipartGeometry
from shapely.geometry.polygon import orient
from shapely.ops import transform, orient, unary_union
from .RepairEntry import RepairEntry

from asf_search.exceptions import ASFWKTError


def validate_wkt(aoi: Union[str, BaseGeometry]) -> Tuple[BaseGeometry, List[RepairEntry]]:
    """
    Param aoi: the WKT string or Shapely Geometry to validate and prepare for the CMR query
    Validates the given area of interest, and returns a validated and simplified WKT string
    returns: The input AOI's CMR ready WKT string
    """
    if isinstance(aoi, str):
        aoi_shape = wkt.loads(aoi)
    else:
        aoi_shape = wkt.loads(aoi.wkt)

    if not aoi_shape.is_valid:
        aoi_shape = _search_wkt_prep(aoi_shape)

        if not aoi_shape.is_valid and not isinstance(aoi_shape, MultiPolygon):
            if isinstance(aoi_shape, Polygon):
                if not aoi_shape.exterior.is_simple:
                    raise ASFWKTError(f'WKT string: \"{aoi_shape.wkt}\" is a self intersecting polygon')

            raise ASFWKTError(f'WKT string: \"{aoi_shape.wkt}\" is not a valid WKT string')

    if aoi_shape.is_empty:
        raise ASFWKTError(f'WKT string: \"{aoi_shape.wkt}\" empty WKT is not a valid AOI')
        
    wrapped, unwrapped, reports = _simplify_geometry(aoi_shape)
    
    return wrapped, unwrapped, [report for report in reports if report != None]


def _search_wkt_prep(shape: BaseGeometry):
    if isinstance(shape, MultiPolygon) :
        output = []
        for geom in shape.geoms:
            output.append(orient(Polygon(geom.exterior)))

        return MultiPolygon(output)
                         
    
    if isinstance(shape, Polygon):
        return orient(Polygon(shape.exterior), sign=1.0)

def _simplify_geometry(geometry: BaseGeometry) -> Tuple[BaseGeometry, List[RepairEntry]]:
    """
    param geometry: AOI Shapely Geometry to be prepped for CMR 
    prepares geometry for CMR by:
        1. Flattening any nested multi-part geometry into single collection
        2. clamping latitude +/-90, unwrapping longitude +/-180, removing coordinate dimensions higher than 2 (lon,lat)
        3. Merging any overlapping shapes
        4. convex-hulling the remainder into a single shape
        4. simplifing until the shape has <= 300 points, with no point closer than 0.00001
        5. Orienting vertices in counter-clockwise winding order
    returns: geometry prepped for CMR
    """
    flattened = _flatten_multipart_geometry(geometry)
    
    merged, merge_report = _merge_overlapping_geometry(flattened)
    convex, convex_report = _get_convex_hull(merged)
    simplified, simplified_report = _simplify_aoi(convex)
    reoriented, reorientation_report = _counter_clockwise_reorientation(simplified)
    wrapped, unwrapped, clamp_report = _get_clamped_and_wrapped_geometry(reoriented)
    
    dimension_report = RepairEntry(
        report_type="'type': 'EXTRA_DIMENSION'", 
        report="'report': Only 2-Dimensional area of interests are supported (lon/lat), higher dimension coordinates will be ignored"
        ) if geometry.has_z else None

    if convex_report != None:
        merge_report = None

    repair_reports = [dimension_report, merge_report, convex_report, *clamp_report, *simplified_report, reorientation_report]   
    for report in repair_reports:
        if report is not None:
            logging.info(f"{report}")

    validated_wrapped = transform(lambda x, y, z=None: tuple([round(x, 14), round(y, 14)]), wrapped)
    validated_unwrapped = transform(lambda x, y, z=None: tuple([round(x, 14), round(y, 14)]), unwrapped)
    return validated_wrapped, validated_unwrapped, repair_reports


def _flatten_multipart_geometry(unflattened_geometry: BaseGeometry) -> BaseGeometry:
    """
    Recursively flattens nested geometric collections, 
    guarantees geometric collections have a depth equal to 1.
    Also ignores any empty shapes in multipart geometry
    """
    def _recurse_nested_geometry(geometry: BaseGeometry) -> List[BaseGeometry]:
        output = []

        if isinstance(geometry, BaseMultipartGeometry):
            for geom in geometry.geoms:
                output.extend(_recurse_nested_geometry(geom))
        elif not geometry.is_empty:
            if isinstance(geometry, Polygon):
                return [Polygon(geometry.exterior)]
            return [geometry]

        return output
    
    flattened = _recurse_nested_geometry(unflattened_geometry)

    return flattened[0] if len(flattened) == 1 else GeometryCollection(flattened)


def _merge_overlapping_geometry(geometry: BaseGeometry) -> Tuple[BaseGeometry, RepairEntry]:
    """
    parameter geometry: geometry to merge
    Performs a unary union overlapping operation of the input geometry, 
    ensuring geometric collections (multipolygon, multipartgeometry, etc)
    are simplied as much as possible before the convex-hull step
    output: merged-overlapping geometry
    """
    merge_report = None

    if isinstance(geometry, BaseMultipartGeometry):
        original_amount = len(geometry.geoms)
        if original_amount == 1:
            return geometry, merge_report

        merged = unary_union(geometry)

        # if there were non-overlapping shapes
        if isinstance(merged, BaseMultipartGeometry):
            unique_shapes = len(merged.geoms)
            merged = orient(unary_union(GeometryCollection([geom.convex_hull for geom in merged.geoms])))
            if isinstance(merged, BaseMultipartGeometry):
                if unique_shapes != len(merged.geoms):
                    merge_report = RepairEntry("'type': 'OVERLAP_MERGE'", f"'report': {unique_shapes - len(merged.geoms)} non-overlapping shapes merged by their convex-hulls")
            else:
                merge_report = RepairEntry("'type': 'OVERLAP_MERGE'", f"'report': {unique_shapes} non-overlapping shapes merged by their convex-hulls")
        else:
            merge_report = RepairEntry("'type': 'OVERLAP_MERGE'", f"'report': Overlapping {original_amount} shapes merged into one")

        return merged, merge_report

    return geometry, merge_report


def _counter_clockwise_reorientation(geometry: Union[Point, LineString, Polygon]):
    """
        param geometry: Shapely geometry to re-orient
        Ensures the geometry coordinates are wound counter-clockwise
        output: counter-clockwise oriented geometry
    """
    reoriented_report = RepairEntry("'type': 'REVERSE'", "'report': Reversed polygon winding order")
    reoriented = orient(geometry)
    
    if isinstance(geometry, Polygon):
        # if the vertice ordering has changed
        if reoriented.exterior.is_ccw != geometry.exterior.is_ccw:
            return reoriented, reoriented_report

    return reoriented, None


def _get_clamped_and_wrapped_geometry(shape: BaseGeometry) -> Tuple[BaseGeometry, List[RepairEntry]]:
    """
    param geometry: Shapely geometry to clamp    
    Clamps geometry to +/-90 latitude and wraps longitude +/-180
    output: clamped shapely geometry
    """
    coords_clamped = 0
    coords_wrapped = 0
    def _clamp_lat(x, y, z=None):
        clamped = _clamp(y)

        if clamped != y:
            nonlocal coords_clamped
            coords_clamped += 1

        return tuple([x, clamped])

    def _wrap_lon(x, y, z=None):
        wrapped = x
        if abs(x) > 180:
            wrapped = (x + 180) % 360 - 180

        if wrapped != x:
            nonlocal coords_wrapped
            coords_wrapped += 1

        return tuple([wrapped, y])

    def _unwrap_lon(x, y, z=None):
        unwrapped = x if x >= 0 else x + 360    # This undoes wrapping

        return tuple([unwrapped, y])


    clamped_lat = transform(_clamp_lat, shape)
    
    wrapped = transform(_wrap_lon, clamped_lat)
    
    if wrapped.bounds[2] - wrapped.bounds[0] > 180:
        unwrapped = transform(_unwrap_lon, wrapped)
    else:
        unwrapped = wrapped
    
    
    clampRepairReport = None
    wrapRepairReport = None

    if coords_clamped > 0:
        clampRepairReport = RepairEntry("'type': 'CLAMP'", f"'report': 'Clamped {coords_clamped} value(s) to +/-90 latitude'")
    if coords_wrapped > 0:
        wrapRepairReport = RepairEntry("'type': 'WRAP'", f"'report': 'Wrapped {coords_wrapped} value(s) to +/-180 longitude'")

    return (wrapped, unwrapped, [clampRepairReport, wrapRepairReport])


def _get_convex_hull(geometry: BaseGeometry) -> Tuple[BaseGeometry, RepairEntry]:
    """
    param geometry: geometry to perform possible convex hull operation on
    If the given geometry is a collection of geometries, creates a convex-hull encompassing said geometry
    output: convex hull of multi-part geometry, or the original single-shaped geometry 
    """
    if geometry.geom_type not in ['MultiPoint', 'MultiLineString', 'MultiPolygon', 'GeometryCollection']:
        return geometry, None
    
    possible_repair = RepairEntry("'type': 'CONVEX_HULL_INDIVIDUAL'", "'report': 'Unconnected shapes: Convex-hulled each INDIVIDUAL shape to merge them together.'")
    return geometry.convex_hull, possible_repair


def _simplify_aoi(shape: Union[Polygon, LineString, Point], 
                  threshold: float = 0.004,
                  max_depth: int = 10,
        ) -> Tuple[Union[Polygon, LineString, Point], List[RepairEntry]]:
    """
    param shape: Shapely geometry to simplify
    param threshold: point proximity threshold to merge nearby points of geometry with
    param max_depth: the current depth of the recursive call, defaults to 10
    Recursively simplifies geometry with increasing threshold, and 
    until there are no more than 300 points
    output: simplified geometry
    """
    repairs = []

    if shape.geom_type == 'Point':
        return shape, repairs

    ### Check for very small shapes and collapse accordingly
    mbr_width = shape.bounds[2] - shape.bounds[0]
    mbr_height = shape.bounds[3] - shape.bounds[1]
    # If both pass, it's a tiny box. Turn it to a point
    if mbr_width <= threshold and mbr_height <= threshold:
        simplified = shape.centroid
        repair = RepairEntry("'type': 'GEOMETRY_SIMPLIFICATION'",
                            f"'report': 'Shape Collapsed to Point: shape of {_get_shape_coords_len(shape)} simplified to {_get_shape_coords_len(simplified)} with proximity threshold of {threshold}'")
        return simplified, [*repairs, repair]
    # If it's a single line segment, it's already as simple as can be. Don't do anything
    elif shape.geom_type == 'LineString' and len(shape.coords) == 2:
        return shape, repairs
    # Else, check if it's slim enough to become a linestring:
    elif mbr_width <= threshold:
        lon = (shape.bounds[2] - shape.bounds[0]) / 2 + shape.bounds[0]
        simplified = LineString([(lon, shape.bounds[1]), (lon, shape.bounds[3])])
        repair = RepairEntry("'type': 'GEOMETRY_SIMPLIFICATION'",
                            f"'report': 'Shape Collapsed to Vertical Line: shape of {_get_shape_coords_len(shape)} simplified to {_get_shape_coords_len(simplified)} with proximity threshold of {threshold}'")
        return simplified, [*repairs, repair]
    elif mbr_height <= threshold:
        lat = (shape.bounds[3] - shape.bounds[1]) / 2 + shape.bounds[1]
        simplified = LineString([(shape.bounds[0], lat), (shape.bounds[2], lat)])
        repair = RepairEntry("'type': 'GEOMETRY_SIMPLIFICATION'",
                            f"'report': 'Shape Collapsed to Horizontal Line: shape of {_get_shape_coords_len(shape)} simplified to {_get_shape_coords_len(simplified)} with proximity threshold of {threshold}'")
        return simplified, [*repairs, repair]

    ### Keep taking away points until it's under 300:
    for simplify_level in range(0, max_depth):
        simplifed = shape.simplify(tolerance=threshold*(1.5**simplify_level))

        coords_length = _get_shape_coords_len(simplifed)
        if _get_shape_coords_len(shape) != coords_length:
            repairs.append(RepairEntry("'type': 'GEOMETRY_SIMPLIFICATION'", f"'report': 'Shape Simplified: shape of {_get_shape_coords_len(shape)} simplified to {coords_length} with proximity threshold of {threshold}'"))

        if coords_length <= 300:
            return simplifed, repairs

    raise ASFWKTError(f"Failed to simplify wkt string: {shape.wkt}")


def _clamp(num):
    """Clamps value between -90 and 90"""
    return max(-90, min(90, num))


def _get_shape_coords_len(geometry: BaseGeometry):
    return len(_get_shape_coords(geometry))


def _get_shape_coords(geometry: BaseGeometry):
    """Returns flattened coordinates of input Shapely geometry"""
    if geometry.geom_type == 'Polygon':
        return list(geometry.exterior.coords[:-1])
    
    if geometry.geom_type == 'LineString':
        return list(geometry.coords)
    
    if geometry.geom_type == 'Point':
        return list(geometry.coords)

    output = []
    
    for geom in geometry.geoms:
        coords = _get_shape_coords(geom)
        output = [*output, *coords]

    return output



import requests
import json

import asf_search.constants

def health(host: str = None) -> dict:
    """
    Checks basic connectivity to and health of the ASF SearchAPI.

    :param host: SearchAPI host, defaults to Production SearchAPI. This option is intended for dev/test purposes.
    :return: Current configuration and status of subsystems
    """

    if host is None:
        host = asf_search.INTERNAL.CMR_HOST
    return json.loads(requests.get(f'https://{host}{asf_search.INTERNAL.CMR_HEALTH_PATH}').text)


import datetime
from typing import Sequence, Tuple, Union
from copy import copy
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.CMR.subquery import build_subqueries
from asf_search.CMR import translate_opts
from asf_search.search.search_generator import get_page, preprocess_opts
from asf_search import INTERNAL


def search_count(        
        absoluteOrbit: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        asfFrame: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        beamMode: Union[str, Sequence[str]] = None,
        beamSwath: Union[str, Sequence[str]] = None,
        campaign: Union[str, Sequence[str]] = None,
        maxDoppler: float = None,
        minDoppler: float = None,
        end: Union[datetime.datetime, str] = None,
        maxFaradayRotation: float = None,
        minFaradayRotation: float = None,
        flightDirection: str = None,
        flightLine: str = None,
        frame: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        granule_list: Union[str, Sequence[str]] = None,
        groupID: Union[str, Sequence[str]] = None,
        insarStackId: str = None,
        instrument: Union[str, Sequence[str]] = None,
        intersectsWith: str = None,
        lookDirection: Union[str, Sequence[str]] = None,
        offNadirAngle: Union[float, Tuple[float, float], Sequence[Union[float, Tuple[float, float]]]] = None,
        platform: Union[str, Sequence[str]] = None,
        polarization: Union[str, Sequence[str]] = None,
        processingDate: Union[datetime.datetime, str] = None,
        processingLevel: Union[str, Sequence[str]] = None,
        product_list: Union[str, Sequence[str]] = None,
        relativeOrbit: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        season: Tuple[int, int] = None,
        start: Union[datetime.datetime, str] = None,
        absoluteBurstID: Union[int, Sequence[int]] = None,
        relativeBurstID: Union[int, Sequence[int]] = None,
        fullBurstID: Union[str, Sequence[str]] = None,
        collections: Union[str, Sequence[str]] = None,
        temporalBaselineDays: Union[str, Sequence[str]] = None,
        operaBurstID: Union[str, Sequence[str]] = None,
        dataset: Union[str, Sequence[str]] = None,
        maxResults: int = None,
        opts: ASFSearchOptions = None,
) -> int:
    # Create a kwargs dict, that's all of the 'not None' items, and merge it with opts:
    kwargs = locals()
    opts = (ASFSearchOptions() if kwargs["opts"] is None else copy(opts))
    del kwargs["opts"]

    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    kw_opts = ASFSearchOptions(**kwargs)

    # Anything passed in as kwargs has priority over anything in opts:
    opts.merge_args(**dict(kw_opts))

    preprocess_opts(opts)

    url = '/'.join(s.strip('/') for s in [f'https://{opts.host}', f'{INTERNAL.CMR_GRANULE_PATH}'])

    count = 0
    for query in build_subqueries(opts):
        translated_opts = translate_opts(query)
        idx = translated_opts.index(('page_size', INTERNAL.CMR_PAGE_SIZE))
        translated_opts[idx] = ('page_size', 0)
        
        response = get_page(session=opts.session, url=url, translated_opts=translated_opts)
        count += response.json()['hits']
    return count



import logging
import pytest
import unittest

from asf_search import ASFProduct, ASFSearchResults, ASFSearchOptions, FileDownloadType
from unittest.mock import patch
from shapely.geometry import shape
from shapely.ops import orient

import requests

def run_test_ASFProduct(product_json):
    if product_json is None:
        product = ASFProduct()
        geojson = product.geojson()
        assert geojson['type'] == 'Feature'
        assert geojson['geometry'] == {'coordinates': None, 'type': 'Polygon'}
        for val in geojson['properties'].values():
            assert val is None

        return

    opts = product_json.pop('opts', None)
    product = ASFProduct(product_json, opts)

    geojson = product.geojson()
    
    if geojson['geometry']['coordinates'] is not None:
        expected_shape = orient(shape(product_json['geometry']))
        output_shape = orient(shape(geojson['geometry'])) 
        assert(output_shape.equals(expected_shape))
    elif product.meta != {}:
        assert product.properties == product_json['properties']
        assert product.geometry == product_json['geometry']

    assert(product.umm == product_json["umm"])
    assert(product.meta == product_json["meta"])

def run_test_stack(reference, pre_processed_stack, processed_stack):
    product = ASFProduct(reference)
    
    with patch('asf_search.baseline_search.search') as search_mock:
        temp = ASFSearchResults([ASFProduct(prod) for prod in pre_processed_stack])
        for idx, prod in enumerate(temp):
            prod.baseline = pre_processed_stack[idx]['baseline']
        search_mock.return_value = temp
        stack = product.stack()


        stack = [
            product for product in stack if product.properties['temporalBaseline'] != None and product.properties['perpendicularBaseline'] != None
            ]

        for(idx, secondary) in enumerate(stack):
            
            if(idx > 0):
                assert(secondary.properties['temporalBaseline'] >= stack[idx - 1].properties['temporalBaseline'])
            
            assert(secondary.properties['temporalBaseline'] == processed_stack[idx]['properties']['temporalBaseline'])
            assert(secondary.properties['perpendicularBaseline'] == processed_stack[idx]['properties']['perpendicularBaseline'])

def run_test_product_get_stack_options(reference, options):
    product = ASFProduct(reference)
    expected_options = dict(ASFSearchOptions(**options))

    product_options = dict(product.get_stack_opts())
    assert product_options == dict(expected_options)

def run_test_ASFProduct_download(reference, filename, filetype, additional_urls):
    product = ASFProduct(reference)
    product.properties['additionalUrls'] = additional_urls
    with patch('asf_search.ASFSession.get') as mock_get:
        resp = requests.Response()
        resp.status_code = 200
        mock_get.return_value = resp
        resp.iter_content = lambda chunk_size: []
            
        with patch('builtins.open', unittest.mock.mock_open()) as m:    
            if filename != None and (
                (filetype == FileDownloadType.ADDITIONAL_FILES and len(additional_urls) > 1) 
                or filetype == FileDownloadType.ALL_FILES
            ):
                with pytest.warns(Warning):
                    product.download('./', filename=filename, fileType=filetype)
            else:
                product.download('./', filename=filename, fileType=filetype)


import inspect
import json
from types import GeneratorType

from asf_search import ASF_LOGGER
from .jsonlite import JSONLiteStreamArray

def results_to_jsonlite2(results):
    ASF_LOGGER.info('started translating results to jsonlite2 format')

    if not inspect.isgeneratorfunction(results) and not isinstance(results, GeneratorType):
        results = [results]
    
    streamer = JSONLite2StreamArray(results)

    for p in json.JSONEncoder(sort_keys=True, separators=(',', ':')).iterencode({'results': streamer}):
        yield p

class JSONLite2StreamArray(JSONLiteStreamArray):
    def getItem(self, p):
         # pre-processing of the result is the same as in the base jsonlite streamer,
         # so use that and then rename/substitute fields
        p = super().getItem(p)
        result = {
            'b': [a.replace(p['granuleName'], '{gn}') for a in p['browse']] if p['browse'] is not None else p['browse'],
            'bm': p['beamMode'],
            'd': p['dataset'],
            'du': p['downloadUrl'].replace(p['granuleName'], '{gn}'),
            'f': p['frame'],
            'fd': p['flightDirection'],
            'fl': p['flightLine'],
            'fn': p['fileName'].replace(p['granuleName'], '{gn}'),
            'fr': p['faradayRotation'], # ALOS
            'gid': p['groupID'].replace(p['granuleName'], '{gn}'),
            'gn': p['granuleName'],
            'i': p['instrument'],
            'in': p['canInSAR'],
            'mn': p['missionName'],
            'o': p['orbit'],
            'on': p['offNadirAngle'], # ALOS
            'p': p['path'],
            'pid': p['productID'].replace(p['granuleName'], '{gn}'),
            'pa': p['pointingAngle'],
            'po': p['polarization'],
            'pt': p['productType'],
            'ptd': p['productTypeDisplay'],
            's': p['sizeMB'],
            'ss': p['stackSize'], # Used for datasets with precalculated stacks
            'st': p['startTime'],
            'stp': p['stopTime'],
            't': p['thumb'].replace(p['granuleName'], '{gn}') if p['thumb'] is not None else p['thumb'],
            'w': p['wkt'],
            'wu': p['wkt_unwrapped'],
            'pge': p['pgeVersion']
        }

        if 'temporalBaseline' in p.keys() or 'perpendicularBaseline' in p.keys():
            result['tb'] = p['temporalBaseline']
            result['pb'] = p['perpendicularBaseline']

        if p.get('burst') is not None: # is a burst product
            result['s1b'] = p['burst']
        
        return result

    def getOutputType(self) -> str:
        return 'jsonlite2'



import logging
from typing import Generator, Union, Sequence, Tuple, List
from copy import copy
from requests.exceptions import HTTPError
from requests import ReadTimeout, Response
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_exponential, wait_fixed
import datetime
import dateparser
import warnings

from asf_search import __version__

from asf_search.ASFSearchResults import ASFSearchResults
from asf_search.ASFSearchOptions import ASFSearchOptions
from asf_search.CMR import build_subqueries, translate_opts
from asf_search.ASFSession import ASFSession
from asf_search.ASFProduct import ASFProduct
from asf_search.exceptions import ASFSearch4xxError, ASFSearch5xxError, ASFSearchError, CMRIncompleteError
from asf_search.constants import INTERNAL
from asf_search.WKT.validate_wkt import validate_wkt
from asf_search.search.error_reporting import report_search_error

def search_generator(        
        absoluteOrbit: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        asfFrame: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        beamMode: Union[str, Sequence[str]] = None,
        beamSwath: Union[str, Sequence[str]] = None,
        campaign: Union[str, Sequence[str]] = None,
        maxDoppler: float = None,
        minDoppler: float = None,
        end: Union[datetime.datetime, str] = None,
        maxFaradayRotation: float = None,
        minFaradayRotation: float = None,
        flightDirection: str = None,
        flightLine: str = None,
        frame: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        granule_list: Union[str, Sequence[str]] = None,
        groupID: Union[str, Sequence[str]] = None,
        insarStackId: str = None,
        instrument: Union[str, Sequence[str]] = None,
        intersectsWith: str = None,
        lookDirection: Union[str, Sequence[str]] = None,
        offNadirAngle: Union[float, Tuple[float, float], Sequence[Union[float, Tuple[float, float]]]] = None,
        platform: Union[str, Sequence[str]] = None,
        polarization: Union[str, Sequence[str]] = None,
        processingDate: Union[datetime.datetime, str] = None,
        processingLevel: Union[str, Sequence[str]] = None,
        product_list: Union[str, Sequence[str]] = None,
        relativeOrbit: Union[int, Tuple[int, int], range, Sequence[Union[int, Tuple[int, int], range]]] = None,
        season: Tuple[int, int] = None,
        start: Union[datetime.datetime, str] = None,
        absoluteBurstID: Union[int, Sequence[int]] = None,
        relativeBurstID: Union[int, Sequence[int]] = None,
        fullBurstID: Union[str, Sequence[str]] = None,
        collections: Union[str, Sequence[str]] = None,
        temporalBaselineDays: Union[str, Sequence[str]] = None,
        operaBurstID: Union[str, Sequence[str]] = None,
        dataset: Union[str, Sequence[str]] = None,
        maxResults: int = None,
        opts: ASFSearchOptions = None,
        ) -> Generator[ASFSearchResults, None, None]:
    # Create a kwargs dict, that's all of the 'not None' items, and merge it with opts:
    kwargs = locals()
    opts = (ASFSearchOptions() if kwargs["opts"] is None else copy(opts))
    del kwargs["opts"]

    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    kw_opts = ASFSearchOptions(**kwargs)

    # Anything passed in as kwargs has priority over anything in opts:
    opts.merge_args(**dict(kw_opts))

    maxResults = opts.pop("maxResults", None)

    if maxResults is not None and \
        (getattr(opts, 'granule_list', False) or getattr(opts, 'product_list', False)):
            raise ValueError("Cannot use maxResults along with product_list/granule_list.")

    preprocess_opts(opts)

    url = '/'.join(s.strip('/') for s in [f'https://{opts.host}', f'{INTERNAL.CMR_GRANULE_PATH}'])
    total = 0
    
    queries = build_subqueries(opts)
    for query in queries:
        translated_opts = translate_opts(query)
        cmr_search_after_header = ""
        subquery_count = 0
        
        while(cmr_search_after_header is not None):
            try:
                items, subquery_max_results, cmr_search_after_header = query_cmr(opts.session, url, translated_opts, subquery_count)
            except (ASFSearchError, CMRIncompleteError) as e:
                message = str(e)
                logging.error(message)
                report_search_error(query, message)
                opts.session.headers.pop('CMR-Search-After', None)
                return
            
            opts.session.headers.update({'CMR-Search-After': cmr_search_after_header})
            last_page = process_page(items, maxResults, subquery_max_results, total, subquery_count, opts)
            subquery_count += len(last_page)
            total += len(last_page)
            last_page.searchComplete = subquery_count == subquery_max_results or total == maxResults
            yield last_page
            
            if last_page.searchComplete:
                if total == maxResults: # the user has as many results as they wanted
                    opts.session.headers.pop('CMR-Search-After', None)
                    return
                else: # or we've gotten all possible results for this subquery
                    cmr_search_after_header = None
        
        opts.session.headers.pop('CMR-Search-After', None)


@retry(reraise=True,
       retry=retry_if_exception_type(CMRIncompleteError),
       wait=wait_fixed(2),
       stop=stop_after_attempt(3),
    )
def query_cmr(session: ASFSession, url: str, translated_opts: dict, sub_query_count: int):
    response = get_page(session=session, url=url, translated_opts=translated_opts)

    items = [ASFProduct(f, session=session) for f in response.json()['items']]
    hits: int = response.json()['hits'] # total count of products given search opts

    # sometimes CMR returns results with the wrong page size
    if len(items) != INTERNAL.CMR_PAGE_SIZE and len(items) + sub_query_count < hits:
        raise CMRIncompleteError(f"CMR returned page of incomplete results. Expected {min(INTERNAL.CMR_PAGE_SIZE, hits - sub_query_count)} results, got {len(items)}")

    return items, hits, response.headers.get('CMR-Search-After', None)
    

def process_page(items: List[ASFProduct], max_results: int, subquery_max_results: int, total: int, subquery_count: int, opts: ASFSearchOptions):
    if max_results is None:
        last_page = ASFSearchResults(items[:min(subquery_max_results - subquery_count, len(items))], opts=opts)
    else:
        last_page = ASFSearchResults(items[:min(max_results - total, len(items))], opts=opts)
    return last_page


@retry(reraise=True,
       retry=retry_if_exception_type(ASFSearch5xxError),
       wait=wait_exponential(multiplier=1, min=3, max=10),  # Wait 2^x * 1 starting with 3 seconds, max 10 seconds between retries
       stop=stop_after_attempt(3),
    )
def get_page(session: ASFSession, url: str, translated_opts: list) -> Response:
    try:
        response = session.post(url=url, data=translated_opts, timeout=INTERNAL.CMR_TIMEOUT)
        response.raise_for_status()
    except HTTPError as exc:
        error_message = f'HTTP {response.status_code}: {response.json()["errors"]}'
        if 400 <= response.status_code <= 499:
            raise ASFSearch4xxError(error_message) from exc
        if 500 <= response.status_code <= 599:
            raise ASFSearch5xxError(error_message) from exc
    except ReadTimeout as exc:
        raise ASFSearchError(f'Connection Error (Timeout): CMR took too long to respond. Set asf constant "CMR_TIMEOUT" to increase. ({url=}, timeout={INTERNAL.CMR_TIMEOUT})') from exc
    
    return response
    

def preprocess_opts(opts: ASFSearchOptions):
    # Repair WKT here so it only happens once, and you can save the result to the new Opts object:
    wrap_wkt(opts=opts)

    # Date/Time logic, convert "today" to the literal timestamp if needed:
    set_default_dates(opts=opts)

    # Platform Alias logic:
    set_platform_alias(opts=opts)


def wrap_wkt(opts: ASFSearchOptions):
    if opts.intersectsWith is not None:
        wrapped, _, __ = validate_wkt(opts.intersectsWith)
        opts.intersectsWith = wrapped.wkt


def set_default_dates(opts: ASFSearchOptions):
    if opts.start is not None and isinstance(opts.start, str):
        opts.start = dateparser.parse(opts.start, settings={'RETURN_AS_TIMEZONE_AWARE': True})
    if opts.end is not None and isinstance(opts.end, str):
        opts.end = dateparser.parse(opts.end, settings={'RETURN_AS_TIMEZONE_AWARE': True})
    # If both are used, make sure they're in the right order:
    if opts.start is not None and opts.end is not None:
        if opts.start > opts.end:
            warnings.warn(f"Start date ({opts.start}) is after end date ({opts.end}). Switching the two.")
            opts.start, opts.end = opts.end, opts.start
    # Can't do this sooner, since you need to compare start vs end:
    if opts.start is not None:
        opts.start = opts.start.strftime('%Y-%m-%dT%H:%M:%SZ')
    if opts.end is not None:
        opts.end = opts.end.strftime('%Y-%m-%dT%H:%M:%SZ')


def set_platform_alias(opts: ASFSearchOptions):
    # Platform Alias logic:
    if opts.platform is not None:
        plat_aliases = {
            # Groups:
            'S1': ['SENTINEL-1A', 'SENTINEL-1B'],
            'SENTINEL-1': ['SENTINEL-1A', 'SENTINEL-1B'],
            'SENTINEL': ['SENTINEL-1A', 'SENTINEL-1B'],
            'ERS': ['ERS-1', 'ERS-2'],
            'SIR-C': ['STS-59', 'STS-68'],
            # Singles / Aliases:
            'R1': ['RADARSAT-1'],
            'E1': ['ERS-1'],
            'E2': ['ERS-2'],
            'J1': ['JERS-1'],
            'A3': ['ALOS'],
            'AS': ['DC-8'],
            'AIRSAR': ['DC-8'],
            'SS': ['SEASAT 1'],
            'SEASAT': ['SEASAT 1'],
            'SA': ['SENTINEL-1A'],
            'SB': ['SENTINEL-1B'],
            'SP': ['SMAP'],
            'UA': ['G-III'],
            'UAVSAR': ['G-III'],
        }
        platform_list = []
        for plat in opts.platform:
            # If it's a key, *replace* it with all the values. Else just add the key:
            if plat.upper() in plat_aliases:
                platform_list.extend(plat_aliases[plat.upper()])
            else:
                platform_list.append(plat)

        opts.platform = list(set(platform_list))



import inspect
import json
from types import GeneratorType
from typing import Tuple
from shapely.geometry import shape
from shapely.ops import transform

from asf_search.CMR.translate import get_additional_fields
from asf_search import ASF_LOGGER, ASFProduct
from asf_search.export.export_translators import ASFSearchResults_to_properties_list

extra_jsonlite_fields = [
    ('processingTypeDisplay', ['AdditionalAttributes', ('Name', 'PROCESSING_TYPE_DISPLAY'), 'Values', 0]),
    ('thumb', ['AdditionalAttributes', ('Name', 'THUMBNAIL_URL'), 'Values', 0]),
    ('faradayRotation', ['AdditionalAttributes', ('Name', 'FARADAY_ROTATION'), 'Values', 0]),
    ('sizeMB', ['DataGranule', 'ArchiveAndDistributionInformation', 0, 'Size']),
    ('flightLine', ['AdditionalAttributes', ('Name', 'FLIGHT_LINE'), 'Values', 0]),
    ('missionName', ['AdditionalAttributes', ('Name', 'MISSION_NAME'), 'Values', 0]),
]

def results_to_jsonlite(results):
    ASF_LOGGER.info('started translating results to jsonlite format')

    if not inspect.isgeneratorfunction(results) and not isinstance(results, GeneratorType):
        results = [results]
    
    streamer = JSONLiteStreamArray(results)
    jsondata = {'results': streamer}

    for p in json.JSONEncoder(indent=2, sort_keys=True).iterencode(jsondata):
        yield p

def unwrap_shape(x, y, z=None):
        x = x if x > 0 else x + 360
        return tuple([x, y])

def get_wkts(geometry) -> Tuple[str, str]:
    wrapped = shape(geometry)
    
    min_lon, max_lon = (wrapped.bounds[0], wrapped.bounds[2])
    
    if max_lon - min_lon > 180:    
        unwrapped = transform(unwrap_shape, wrapped)
    else:
        unwrapped = wrapped

    return wrapped.wkt, unwrapped.wkt

class JSONLiteStreamArray(list):
    def __init__(self, results):
        self.results = results

        # need to make sure we actually have results so we can intelligently set __len__, otherwise
        # iterencode behaves strangely and will output invalid json
        self.len = 1

    def __iter__(self):
        return self.streamDicts()

    def __len__(self):
        return self.len

    def get_additional_output_fields(self, product: ASFProduct):
        umm = product.umm
        
        additional_fields = {}
        for key, path in extra_jsonlite_fields:
            additional_fields[key] = get_additional_fields(umm, *path)

        if product.properties['platform'].upper() in ['ALOS', 'RADARSAT-1', 'JERS-1', 'ERS-1', 'ERS-2']:
            insarGrouping = get_additional_fields(umm, *['AdditionalAttributes', ('Name', 'INSAR_STACK_ID'), 'Values', 0])
            
            if insarGrouping not in [None, 0, '0', 'NA', 'NULL']:
                additional_fields['canInsar'] = True
                additional_fields['insarStackSize'] = get_additional_fields(umm, *['AdditionalAttributes', ('Name', 'INSAR_STACK_SIZE'), 'Values', 0])
            else:
                additional_fields['canInsar'] = False
        else:
            additional_fields['canInsar'] = product.baseline is not None

        additional_fields['geometry'] = product.geometry
        
        return additional_fields

    def streamDicts(self):
        
        completed = False
        for page_idx, page in enumerate(self.results):
            ASF_LOGGER.info(f"Streaming {len(page)} products from page {page_idx}")
            completed = page.searchComplete
            
            yield from [self.getItem(p) for p in ASFSearchResults_to_properties_list(page, self.get_additional_output_fields) if p is not None]

        if not completed:
            ASF_LOGGER.warn('Failed to download all results from CMR')

        ASF_LOGGER.info(f"Finished streaming {self.getOutputType()} results")
    
    def getItem(self, p):
        for i in p.keys():
            if p[i] == 'NA' or p[i] == '':
                p[i] = None
        try:
            if float(p['offNadirAngle']) < 0:
                p['offNadirAngle'] = None
        except TypeError:
            pass

        try:
            if float(p['pathNumber']) < 0:
                p['pathNumber'] = None
        except TypeError:
            pass

        try:
            if p['groupID'] is None:
                p['groupID'] = p['sceneName']
        except TypeError:
            pass

        try:
            p['sizeMB'] = float(p['sizeMB'])
        except TypeError:
            pass

        try:
            p['pathNumber'] = int(p['pathNumber'])
        except TypeError:
            pass

        try:
            p['frameNumber'] = int(p['frameNumber'])
        except TypeError:
            pass

        try:
            p['orbit'] = int(p['orbit'])
        except TypeError:
            pass

        wrapped, unwrapped = get_wkts(p['geometry'])
        result = {
            'beamMode': p['beamModeType'],
            'browse': [] if p.get('browse') is None else p.get('browse'),
            'canInSAR': p['canInsar'],
            'dataset': p['platform'],
            'downloadUrl': p['url'],
            'faradayRotation': p['faradayRotation'], # ALOS
            'fileName': p['fileName'],
            'flightDirection': p['flightDirection'],
            'flightLine': p['flightLine'],
            'frame': p['frameNumber'],
            'granuleName': p['sceneName'],
            'groupID': p['groupID'],
            'instrument': p['sensor'],
            'missionName': p['missionName'],
            'offNadirAngle': str(p['offNadirAngle']) if p['offNadirAngle'] is not None else None, # ALOS
            'orbit': [str(p['orbit'])],
            'path': p['pathNumber'],
            'polarization': p['polarization'],
            'pointingAngle': p['pointingAngle'],
            'productID': p['fileID'],
            'productType': p['processingLevel'],
            'productTypeDisplay': p['processingTypeDisplay'],
            'sizeMB': p['sizeMB'],
            'stackSize': p.get('insarStackSize'), # Used for datasets with precalculated stacks
            'startTime': p['startTime'],
            'stopTime': p['stopTime'],
            'thumb': p['thumb'],
            'wkt': wrapped,
            'wkt_unwrapped': unwrapped,
            'pgeVersion': p['pgeVersion']
        }
        
        for key in result.keys():
            if result[key] in [ 'NA', 'NULL']:
                result[key] = None

        if 'temporalBaseline' in p.keys() or 'perpendicularBaseline' in p.keys():
            result['temporalBaseline'] = p['temporalBaseline']
            result['perpendicularBaseline'] = p['perpendicularBaseline']

        if p.get('processingLevel') == 'BURST': # is a burst product
            result['burst'] = p['burst']

        return result

    def getOutputType(self) -> str:
        return 'jsonlite'
    